﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Objects.DataClasses;
using System.Data.Objects;
using System.Data.Entity;
using System.Data;
using Nido.Common.Exceptions;
using System.Data.SqlClient;
using System.Linq.Expressions;
using System.Data.Entity.Infrastructure;
using AdminPortal.Common;

namespace Nido.Common.BackEnd
{
    /// <summary>
    /// Generic repositor class. 
    /// </summary>
    /// <typeparam name="TContext"></typeparam>
    /// <typeparam name="Entity"></typeparam>
    public class DataRepository<TContext, Entity> : IDisposable
        where TContext : BaseObjectConext, IBaseObjectConext
        where Entity : class, IBaseObject
    {
        /// <summary>
        /// Default constructor
        /// </summary>
        public DataRepository()
        { }

        /// <summary>
        ///  Cached ObjectSets so changes persist
        /// </summary>
        protected Dictionary<string, object> CachedObjects = new Dictionary<string, object>();

        /// <summary>
        /// Generic Get object set method
        /// </summary>
        /// <typeparam name="TEntity"></typeparam>
        /// <returns></returns>
        [Obsolete("GetObjectSet<TEntity>() is deprecated, please use GetObjectSet() instead.")]
        protected DbSet<TEntity> GetObjectSet<TEntity>() where TEntity : class
        {
            var fulltypename = typeof(TEntity).AssemblyQualifiedName;
            if (fulltypename == null)
                throw new ArgumentException("Invalid Type passed to GetObjectSet!");
            if (!CachedObjects.ContainsKey(fulltypename))
            {
                var objectset = _context.Set<TEntity>();
                CachedObjects.Add(fulltypename, objectset);
            }
            return CachedObjects[fulltypename] as DbSet<TEntity>;
        }

        /// <summary>
        /// Targetted Get object set method. 
        /// By default this will get the object set of the corresponding entity object.
        /// </summary>
        /// <returns>DbSet of the entity</returns>
        protected DbSet<Entity> GetObjectSet()
        {
            /*var fulltypename = typeof(Entity).AssemblyQualifiedName;
            if (fulltypename == null)
                throw new ArgumentException("Invalid Type passed to GetObjectSet!");
            if (_entitySet == null)
            {
                _entitySet = _context.Set<Entity>();
                // CachedObjects.Add(fulltypename, objectset);
            }
            return _entitySet;
            //return CachedObjects[fulltypename] as DbSet<Entity>;*/
            _entitySet = new ContextAgentFactory<TContext, Entity>().GetContext(context).GetObjectSet();
            return _entitySet;
        }

        private TContext context;

        /// <summary>
        /// The generic DbContext
        /// </summary>
        protected TContext _context
        {
            get { return new ContextAgentFactory<TContext, Entity>().GetContext(context).GetContext(); }
            set { context = value; }
        }

        /// <summary>
        /// Generic DbSet object
        /// </summary>
        protected DbSet<Entity> _entitySet;

        /// <summary>
        /// Include table name queue
        /// </summary>
        protected Queue<string> _includes = new Queue<string>();

        /// <summary>     
        /// Constructor that takes a context     
        /// </summary>     
        /// <param name="context">An established data context</param>     
        public DataRepository(TContext context)
        {
            _context = context;
        }

        /// <summary>     
        /// Constructor that takes a connection string and an EDMX name     
        /// </summary>     
        /// <param name="connectionString">The connection string</param>     
        /// <param name="edmxName">The name of the EDMX so we can build an Entity Connection string</param>     
        public DataRepository(string connectionString, string edmxName)
        {
            var entityConnection = String.Format("metadata=res://*/{0}.csdl|res://*/{0}.ssdl|res://*/{0}.msl;provider=System.Data.SqlClient;provider connection string=", edmxName);
            // append the database connection string and save           
            entityConnection = entityConnection + "\"" + connectionString + "\"";
            var targetType = typeof(TContext); var ctx = Activator.CreateInstance(targetType, entityConnection);
            _context = (TContext)ctx;
        }

        /// <summary>
        /// Generic fetch method
        /// </summary>
        /// <typeparam name="TEntity">Type of the entity to be fetch</typeparam>
        /// <returns>DbSet value of the fetched entity</returns>
        [Obsolete("Fetch<TEntity>() is deprecated, please use Fetch() instead.")]
        protected DbSet<TEntity> Fetch<TEntity>() where TEntity : class
        {
            return GetObjectSet<TEntity>();
        }

        //public DbSet<Entity> Fetch() 
        //{
        //    return GetObjectSet();
        //}

        /// <summary>
        /// Generic get all method
        /// </summary>
        /// <typeparam name="TEntity">Entity type</typeparam>
        /// <returns>Enumeralbe entity set</returns>
        [Obsolete("GetAll<TEntity>() is deprecated, please use GetAll() instead.")]
        protected IEnumerable<TEntity> GetAll<TEntity>() where TEntity : class
        {
            return GetObjectSet<TEntity>().ToList().AsEnumerable();
        }

        /// <summary>
        /// Get all the record from the database. Inlcude Supported 
        /// </summary>
        /// <returns></returns>
        protected IEnumerable<Entity> GetAllNew()
        {
            try
            {
                if (_includes.Count == 0)
                    return GetObjectSet().AsEnumerable();
                else
                {
                    DbQuery<Entity> entityQuery = GetObjectSet().Include(_includes.Dequeue());
                    while (_includes.Count > 0)
                        entityQuery = entityQuery.Include(_includes.Dequeue());
                    return entityQuery.AsEnumerable();
                }
            }
            catch (ArgumentException e)
            {
                throw new ExceptionDataError(e, ExceptionBase.GetRefId());
            }
            catch (Exception e)
            {
                throw new ExceptionCritical(e, ExceptionBase.GetRefId());
            }
        }

        /// <summary>
        /// Get all the matching record from the database. Inlcude Supported 
        /// </summary>
        /// <returns></returns>
        protected IEnumerable<Entity> GetAllNew(Expression<Func<Entity, Boolean>> predicate)
        {
            try
            {
                if (_includes.Count == 0)
                    return GetObjectSet().Where(predicate).AsEnumerable();
                else
                {
                    DbQuery<Entity> entityQuery = GetObjectSet().Include(_includes.Dequeue());
                    while (_includes.Count > 0)
                        entityQuery = entityQuery.Include(_includes.Dequeue());
                    return entityQuery.Where(predicate).AsEnumerable();
                }
            }
            catch (ArgumentException e)
            {
                throw new ExceptionDataError(e, ExceptionBase.GetRefId());
            }
            catch (Exception e)
            {
                throw new ExceptionCritical(e, ExceptionBase.GetRefId());
            }
        }

        /// <summary>
        /// Find the matching element from the list of database records
        /// </summary>
        /// <typeparam name="TEntity">Type of the entity to be retreived</typeparam>
        /// <param name="predicate">search query to be matched when finding</param>
        /// <returns></returns>
        [Obsolete("Find<TEntity>(Func<TEntity, bool> predicate) is deprecated, please use just Find(Func<Entity, bool> predicate) instead.")]
        public IEnumerable<TEntity> Find<TEntity>(Expression<Func<TEntity, Boolean>> predicate) where TEntity : class
        {
            return GetObjectSet<TEntity>().Where(predicate);
        }

        //public IEnumerable<Entity> Find(Func<Entity, bool> predicate) 
        //{
        //    return GetObjectSet().Where(predicate);
        //}

        /// <summary>
        /// Retrieve a single object. This method is optimized to use with the primary key.
        /// </summary>
        /// <typeparam name="TEntity">Entity type to be retrieved</typeparam>
        /// <param name="predicate">search query</param>
        /// <returns></returns>
        [Obsolete("GetSingle<TEntity>(Func<TEntity, bool> predicate) is deprecated, please use just GetSingle(Func<Entity, bool> predicate) instead.")]
        protected TEntity GetSingle<TEntity>(Expression<Func<TEntity, Boolean>> predicate) where TEntity : class
        {
            return GetObjectSet<TEntity>().Single(predicate);
        }

        /// <summary>
        /// Select a single matching record. Include Supported. 
        /// If you are selecting by primary key, GetSingleNew is semantically the correct methods to call 
        /// and are usually quite fast.
        /// </summary>
        /// <param name="predicate"></param>
        /// <returns></returns>
        public Entity GetSingleNew(Expression<Func<Entity, Boolean>> predicate)
        {
            if (_includes.Count == 0)
                return GetObjectSet().SingleOrDefault(predicate);
            else
            {
                DbQuery<Entity> entityQuery = GetObjectSet().Include(_includes.Dequeue());
                while (_includes.Count > 0)
                    entityQuery = entityQuery.Include(_includes.Dequeue());
                return entityQuery.SingleOrDefault(predicate);
            }
        }

        /// <summary>
        /// Get the first element from the matching query
        /// </summary>
        /// <typeparam name="TEntity">type of the entity to be retrieved</typeparam>
        /// <param name="predicate">matching query</param>
        /// <returns></returns>
        [Obsolete("GetFirst<TEntity>(Func<TEntity, bool> predicate) is deprecated, please use just GetFirst(Func<Entity, bool> predicate) instead.")]
        protected TEntity GetFirst<TEntity>(Expression<Func<TEntity, Boolean>> predicate) where TEntity : class
        {
            return GetObjectSet<TEntity>().First(predicate);
        }

        /// <summary>
        /// Get the first matching record. Include Supported.
        /// Where returns an IQueryable of the same source, equivalent to the corresponding SQL where clause.
        /// </summary>
        /// <param name="predicate"></param>
        /// <returns></returns>
        public Entity GetFirstNew(Expression<Func<Entity, Boolean>> predicate)
        {
            if (_includes.Count == 0)
                return GetObjectSet().Where(predicate).FirstOrDefault();
            else
            {
                DbQuery<Entity> entityQuery = GetObjectSet().Include(_includes.Dequeue());
                while (_includes.Count > 0)
                    entityQuery = entityQuery.Include(_includes.Dequeue());

                IEnumerable<Entity> list = entityQuery.Where(predicate);
                return list.FirstOrDefault();
            }
        }

        /// <summary>
        /// Add a element to the database
        /// </summary>
        /// <typeparam name="TEntity">entity type to be retrieved from the database</typeparam>
        /// <param name="entity">Enetiy object to be add to the database</param>
        [Obsolete("Add<TEntity>(TEntity entity) is deprecated, please use just Add(Entity entity) instead.")]
        protected void Add<TEntity>(TEntity entity) where TEntity : class
        {
            if (entity == null)
                throw new ArgumentException("Cannot add a null entity"); GetObjectSet<TEntity>().Add(entity);
        }

        /// <summary>
        /// Add object to the database.
        /// This is the recomended method.
        /// </summary>
        /// <param name="entity">The entity object to be added to the database</param>
        /// <remarks>Save method is not called so you have to call it manually. 
        /// If you want to update multiple records and do the save at the end 
        /// then it is recomeded to use methods with 'New' Suffix 
        /// and call the Save method at the end.</remarks>
        protected void AddNew(Entity entity)
        {
            if (entity == null)
                throw new ArgumentException("Cannot add a null entity"); GetObjectSet().Add(entity);
        }

        /// <summary>
        /// Delete a record from the database. This will do a cascade 
        /// delete unless you have implement the CanDelete property of 
        /// the respective enetity object.
        /// </summary>
        /// <typeparam name="TEntity">type of the entity to be deleted</typeparam>
        /// <param name="entity">entity object to be used when finding the record to be deleted</param>
        [Obsolete("Delete<TEntity>(TEntity entity) is deprecated, please use just Delete(Entity entity) instead.")]
        protected void Delete<TEntity>(TEntity entity) where TEntity : class
        {
            if (entity == null) throw new ArgumentException("Cannot delete a null entity");
            this.Attach(entity);
            GetObjectSet<TEntity>().Remove(entity);
        }

        /// <summary>
        /// Delete a record from the database. This will do a cascade 
        /// delete unless you have implement the CanDelete property of 
        /// the respective enetity object.
        /// This is the recomended method.
        /// </summary>
        /// <param name="entity">Entity object to be deleted</param>
        /// <remarks>Save method is not called so you have to call it manually. 
        /// If you want to update multiple records and do the save at the end 
        /// then it is recomeded to use methods with 'New' Suffix 
        /// and call the Save method at the end.</remarks> 
        protected void DeleteNew(Entity entity)
        {
            if (entity == null) throw new ArgumentException("Cannot delete a null entity");
            this.AttachNew(entity);
            GetObjectSet().Remove(entity);
        }

        /// <summary>
        /// Generic attach method
        /// </summary>
        /// <typeparam name="TEntity"></typeparam>
        /// <param name="entity"></param>
        [Obsolete("Attach<TEntity>(TEntity entity) is deprecated, please use just Attach(Entity entity) instead.")]
        protected void Attach<TEntity>(TEntity entity) where TEntity : class
        {
            if (entity == null) throw new ArgumentException("Cannot attach a null entity");
            GetObjectSet<TEntity>().Attach(entity);
        }

        /// <summary>
        /// Attach a entity to the database.
        /// Attaching is required if you do custom update of a object. The generic update method does this internally.
        /// This is the recomended method.
        /// </summary>
        /// <param name="entity">entity to be attached</param>
        protected void AttachNew(Entity entity)
        {
            if (entity == null) throw new ArgumentException("Cannot attach a null entity");
            GetObjectSet().Attach(entity);
        }

        /// <summary>
        /// Generic update method. This will fully update the values of the passing object to the database. No partial update is supported.
        /// </summary>
        /// <typeparam name="TEntity">The type of the generic object to be updated</typeparam>
        /// <param name="entity">Entity object to be update to the database.</param>
        /// <returns></returns>
        [Obsolete("Update<TEntity>(TEntity entity) is deprecated, please use just Update(Entity entity) instead.")]
        protected bool Update<TEntity>(TEntity entity) where TEntity : class
        {
            this.Attach<TEntity>(entity);
            // Copy over new values
            _context.Entry(entity).CurrentValues.SetValues(entity);
            // Set the state to modified.
            _context.Entry(entity).State = System.Data.EntityState.Modified;
            return true;
        }

        /// <summary>
        /// Update method. 
        /// This will fully update the values of the passing object to the database. 
        /// No partial update is supported.
        /// Save method is not called.
        /// </summary>
        /// <param name="entity">The entity object to be updated</param>
        /// <returns></returns>
        /// <remarks>Save method is not called so you have to call it manually. 
        /// If you want to update multiple records and do the save at the end 
        /// then it is recomeded to use methods with 'New' Suffix 
        /// and call the Save method at the end.</remarks>
        protected bool UpdateNew(Entity entity)
        {
            this.AttachNew(entity);
            // Copy over new values
            _context.Entry(entity).CurrentValues.SetValues(entity);
            // Set the state to modified.
            _context.Entry(entity).State = System.Data.EntityState.Modified;
            return true;
        }

        /// <summary>
        /// Get the count of records that matches the predicate
        /// </summary>
        /// <param name="predicate">matching predicate</param>
        /// <returns></returns>
        public int CountNew(Expression<Func<Entity, Boolean>> predicate)
        {
            return GetObjectSet().Count(predicate);
        }

        /// <summary>
        /// Get the count of all the records 
        /// </summary>
        /// <returns></returns>
        public int CountNew()
        {
            return GetObjectSet().Count();
        }

        /// <summary>
        /// DB save method. This method is called internally. 
        /// But if you want to update multiple tables in a same call then you suppose to handle it manually.
        /// </summary>
        /// <returns>Save status. The number of records updated.</returns>
        /// <exception cref="ExceptionDataError"></exception>
        /// <exception cref="ExceptionRepeatable"></exception>
        /// <exception cref="ExceptionNoneRepeatable"></exception>        
        /// <exception cref="ExceptionCritical"></exception>
        public int SaveChanges()
        {
            try
            {
                int i = _context.SaveChanges();
                return i;
            }
            catch (ArgumentException e)
            {
                throw new ExceptionDataError(e, ExceptionBase.GetRefId());
            }
            catch (OptimisticConcurrencyException e)
            {
                throw new ExceptionRepeatable(e);
            }
            catch (UpdateException e)
            {
                throw new ExceptionDataError(e, ExceptionBase.GetRefId());
            }
            catch (SqlException e)
            {
                throw new ExceptionNoneRepeatable(e, ExceptionBase.GetRefId());
            }
            catch (Exception e)
            {
                throw new ExceptionCritical(e, ExceptionCritical.GetRefId());
            }
        }

        #region IDisposable Members

        private bool disposedValue;
        /// <summary>
        /// Dispose method
        /// </summary>
        /// <param name="disposing"></param>
        protected void Dispose(bool disposing)
        {
            if (!this.disposedValue)
            {
                if (disposing)
                {
                    // dispose managed state here if required             }            
                    // dispose unmanaged objects and set large fields to null     
                    if (_context != null)
                        _context.Dispose();
                } this.disposedValue = true;
            }
        }
        /// <summary>
        /// Dispose method
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion
    }
}

